@@ -347,7 +347,7 @@ int cram_external_decode_int(cram_slice *slice, cram_codec *c,
     } else {
 	for (i = 0; i < slice->hdr->num_blocks; i++) {
 	    b = slice->block[i];
-	    if (b->content_type == EXTERNAL &&
+	    if (b && b->content_type == EXTERNAL &&
 		b->content_id == c->external.content_id) {
 		break;
 	    }
@@ -378,7 +378,7 @@ int cram_external_decode_char(cram_slice *slice, cram_codec *c,
     } else {
 	for (i = 0; i < slice->hdr->num_blocks; i++) {
 	    b = slice->block[i];
-	    if (b->content_type == EXTERNAL &&
+	    if (b && b->content_type == EXTERNAL &&
 		b->content_id == c->external.content_id) {
 		break;
 	    }
@@ -395,14 +395,27 @@ int cram_external_decode_char(cram_slice *slice, cram_codec *c,
     return 0;
 }
 
-int cram_external_decode_block(cram_slice *slice, cram_codec *c,
-			      cram_block *in, char *out_,
-			      int *out_size) {
-    int i;
+static int cram_external_decode_block2(cram_slice *slice, cram_codec *c,
+				      cram_block *in, char *out_,
+				      int *out_size) {
     char *cp;
-    cram_block *b = NULL;
+    cram_block *b = c->external.b;
     cram_block *out = (cram_block *)out_;
 
+    cp = cram_extract_block(b, *out_size);
+    if (!cp)
+	return -1;
+
+    BLOCK_APPEND(out, cp, *out_size);
+    return 0;
+}
+
+static int cram_external_decode_block(cram_slice *slice, cram_codec *c,
+				       cram_block *in, char *out_,
+				       int *out_size) {
+    int i;
+    cram_block *b = NULL;
+
     /* Find the external block */
     if (slice->block_by_id) {
 	if (!(b = slice->block_by_id[c->external.content_id]))
@@ -410,7 +423,7 @@ int cram_external_decode_block(cram_slice *slice, cram_codec *c,
     } else {
 	for (i = 0; i < slice->hdr->num_blocks; i++) {
 	    b = slice->block[i];
-	    if (b->content_type == EXTERNAL &&
+	    if (b && b->content_type == EXTERNAL &&
 		b->content_id == c->external.content_id) {
 		break;
 	    }
@@ -419,12 +432,10 @@ int cram_external_decode_block(cram_slice *slice, cram_codec *c,
 	    return -1;
     }
 
-    cp = cram_extract_block(b, *out_size);
-    if (!cp)
-	return -1;
+    c->external.b = b;
+    c->decode = cram_external_decode_block2;
 
-    BLOCK_APPEND(out, cp, *out_size);
-    return 0;
+    return c->decode(slice, c, in, out_, out_size);
 }
 
 void cram_external_decode_free(cram_codec *c) {
@@ -463,14 +474,20 @@ cram_codec *cram_external_decode_init(char *data, int size,
     return c;
 }
 
-int cram_external_encode(cram_slice *slice, cram_codec *c,
-			cram_block *out, char *in, int in_size) {
+int cram_external_encode_int(cram_slice *slice, cram_codec *c,
+			     cram_block *out, char *in, int in_size) {
     uint32_t *i32 = (uint32_t *)in;
 
     itf8_put_blk(out, *i32);
     return 0;
 }
 
+int cram_external_encode_char(cram_slice *slice, cram_codec *c,
+			      cram_block *out, char *in, int in_size) {
+    BLOCK_APPEND(out, in, in_size);
+    return 0;
+}
+
 void cram_external_encode_free(cram_codec *c) {
     if (!c)
 	return;
@@ -508,7 +525,12 @@ cram_codec *cram_external_encode_init(cram_stats *st,
 	return NULL;
     c->codec = E_EXTERNAL;
     c->free = cram_external_encode_free;
-    c->encode = cram_external_encode;
+    if (option == E_INT || option == E_LONG)
+	c->encode = cram_external_encode_int;
+    else if (option == E_BYTE_ARRAY || option == E_BYTE)
+	c->encode = cram_external_encode_char;
+    else
+	abort();
     c->store = cram_external_encode_store;
 
     c->e_external.content_id = (size_t)dat;
@@ -528,7 +550,7 @@ int cram_beta_decode_int(cram_slice *slice, cram_codec *c, cram_block *in, char
 	    out_i[i] = get_bits_MSB(in, c->beta.nbits) - c->beta.offset;
     } else {
 	for (i = 0, n = *out_size; i < n; i++)
-	    out_i[i] = 0;
+	    out_i[i] = -c->beta.offset;
     }
 
     return 0;
@@ -542,7 +564,7 @@ int cram_beta_decode_char(cram_slice *slice, cram_codec *c, cram_block *in, char
 	    out[i] = get_bits_MSB(in, c->beta.nbits) - c->beta.offset;
     } else {
 	for (i = 0, n = *out_size; i < n; i++)
-	    out[i] = 0;
+	    out[i] = -c->beta.offset;
     }
 
     return 0;
@@ -646,7 +668,7 @@ cram_codec *cram_beta_encode_init(cram_stats *st,
 	c->encode = cram_beta_encode_char;
     c->store  = cram_beta_encode_store;
 
-    if (dat) {
+    if (dat) { // FIXME, supply external ID separate?
 	min_val = ((int *)dat)[0];
 	max_val = ((int *)dat)[1];
     } else {
@@ -1497,28 +1519,12 @@ cram_codec *cram_byte_array_len_encode_init(cram_stats *st,
  * ---------------------------------------------------------------------------
  * BYTE_ARRAY_STOP
  */
-int cram_byte_array_stop_decode_char(cram_slice *slice, cram_codec *c,
-				     cram_block *in, char *out,
-				     int *out_size) {
-    int i;
-    cram_block *b = NULL;
+static int cram_byte_array_stop_decode_char2(cram_slice *slice, cram_codec *c,
+					     cram_block *in, char *out,
+					     int *out_size) {
+    cram_block *b = c->byte_array_stop.b;
     char *cp, ch;
 
-    if (slice->block_by_id) {
-	if (!(b = slice->block_by_id[c->byte_array_stop.content_id]))
-	    return -1;
-    } else {
-	for (i = 0; i < slice->hdr->num_blocks; i++) {
-	    b = slice->block[i];
-	    if (b->content_type == EXTERNAL &&
-		b->content_id == c->byte_array_stop.content_id) {
-		break;
-	    }
-	}
-	if (i == slice->hdr->num_blocks || !b)
-	    return -1;
-    }
-
     if (b->idx >= b->uncomp_size)
 	return -1;
 
@@ -1536,13 +1542,42 @@ int cram_byte_array_stop_decode_char(cram_slice *slice, cram_codec *c,
     return 0;
 }
 
+static int cram_byte_array_stop_decode_char(cram_slice *slice, cram_codec *c,
+					    cram_block *in, char *out,
+					    int *out_size) {
+    int i;
+    cram_block *b = NULL;
+
+    if (slice->block_by_id) {
+	if (!(b = slice->block_by_id[c->byte_array_stop.content_id]))
+	    return -1;
+    } else {
+	for (i = 0; i < slice->hdr->num_blocks; i++) {
+	    b = slice->block[i];
+	    if (b && b->content_type == EXTERNAL &&
+		b->content_id == c->byte_array_stop.content_id) {
+		break;
+	    }
+	}
+	if (i == slice->hdr->num_blocks || !b)
+	    return -1;
+    }
+
+    if (b->idx >= b->uncomp_size)
+	return -1;
+
+    c->byte_array_stop.b = b;
+    c->decode = cram_byte_array_stop_decode_char2;
+
+    return c->decode(slice, c, in, out, out_size);
+}
+
 int cram_byte_array_stop_decode_block(cram_slice *slice, cram_codec *c,
 				      cram_block *in, char *out_,
 				      int *out_size) {
-    int space = 256;
     cram_block *b = NULL;
     cram_block *out = (cram_block *)out_;
-    char *cp, ch, *out_cp, *cp_end, *out_end;
+    char *cp, *out_cp, *cp_end;
     char stop;
 
     if (slice->block_by_id) {
@@ -1552,7 +1587,7 @@ int cram_byte_array_stop_decode_block(cram_slice *slice, cram_codec *c,
 	int i;
 	for (i = 0; i < slice->hdr->num_blocks; i++) {
 	    b = slice->block[i];
-	    if (b->content_type == EXTERNAL &&
+	    if (b && b->content_type == EXTERNAL &&
 		b->content_id == c->byte_array_stop.content_id) {
 		break;
 	    }
@@ -1767,3 +1802,45 @@ cram_codec *cram_encoder_init(enum cram_encoding codec,
 	abort();
     }
 }
+
+/*
+ * Returns the content_id used by this codec, also in id2 if byte_array_len.
+ * Returns -1 for the CORE block and -2 for unneeded.
+ * id2 is only filled out for BYTE_ARRAY_LEN which uses 2 codecs.
+ */
+int cram_codec_to_id(cram_codec *c, int *id2) {
+    int bnum1, bnum2 = -2;
+
+    switch (c->codec) {
+    case E_HUFFMAN:
+	bnum1 = c->huffman.ncodes == 1 ? -2 : -1;
+	break;
+    case E_GOLOMB:
+    case E_BETA:
+    case E_SUBEXP:
+    case E_GOLOMB_RICE:
+    case E_GAMMA:
+	bnum1 = -1;
+	break;
+    case E_EXTERNAL:
+	bnum1 = c->external.content_id;
+	break;
+    case E_BYTE_ARRAY_LEN:
+	bnum1 = cram_codec_to_id(c->byte_array_len.len_codec, NULL);
+	bnum2 = cram_codec_to_id(c->byte_array_len.value_codec, NULL);
+	break;
+    case E_BYTE_ARRAY_STOP:
+	bnum1 = c->byte_array_stop.content_id;
+	break;
+    case E_NULL:
+	bnum1 = -2;
+	break;
+    default:
+	fprintf(stderr, "Unknown codec type %d\n", c->codec);
+	bnum1 = -1;
+    }
+
+    if (id2)
+	*id2 = bnum2;
+    return bnum1;
+}
