@@ -1,3 +1,36 @@
+/*
+ * Copyright (c) 2007-2010, 2013 Genome Research Ltd.
+ * Author(s): James Bonfield
+ * 
+ * Redistribution and use in source and binary forms, with or without 
+ * modification, are permitted provided that the following conditions are met:
+ * 
+ *    1. Redistributions of source code must retain the above copyright notice,
+ *       this list of conditions and the following disclaimer.
+ * 
+ *    2. Redistributions in binary form must reproduce the above
+ *       copyright notice, this list of conditions and the following
+ *       disclaimer in the documentation and/or other materials provided
+ *       with the distribution.
+ * 
+ *    3. Neither the names Genome Research Ltd and Wellcome Trust Sanger
+ *    Institute nor the names of its contributors may be used to endorse
+ *    or promote products derived from this software without specific
+ *    prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY GENOME RESEARCH LTD AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GENOME RESEARCH
+ * LTD OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
+
 /*
  * This code implements "interlaced deflate", and is *based* on a
  * simplistic implementation of the Deflate algorithm.
@@ -540,6 +573,7 @@ huffman_codes_t *calc_bit_lengths(unsigned char *data, int len,
 	nodes[ncodes].count = eof;
 	nodes[ncodes].parent = NULL;
 	n2[ncodes] = &nodes[ncodes];
+	ncodes++;
     }
 
     /* All 256 chars existing at a minimal level */
@@ -549,6 +583,7 @@ huffman_codes_t *calc_bit_lengths(unsigned char *data, int len,
 	    nodes[ncodes].count = hist[i];
 	    nodes[ncodes].parent = NULL;
 	    n2[ncodes] = &nodes[ncodes];
+	    ncodes++;
 	}
     } else {
 	/* Only include non-zero symbols */
@@ -559,6 +594,7 @@ huffman_codes_t *calc_bit_lengths(unsigned char *data, int len,
 	    nodes[ncodes].count = hist[i];
 	    nodes[ncodes].parent = NULL;
 	    n2[ncodes] = &nodes[ncodes];
+	    ncodes++;
 	}
     }
 
