@@ -45,7 +45,8 @@
 
 #include <io_lib/cram.h>
 
-void DumpMap2(cram_map **ma, FILE *fp, char *prefix, char *data) {
+void DumpMap2(cram_map **ma, FILE *fp, char *prefix, char *data,
+	      HashTable *ds_h) {
     int i, j, k;
     for (i = 0; i < CRAM_MAP_HASH; i++) {
 	cram_map *m;
@@ -60,6 +61,17 @@ void DumpMap2(cram_map **ma, FILE *fp, char *prefix, char *data) {
 		printf(j ? ", %d" : "%d", (unsigned char)data[k]);
 	    }
 	    printf("}\n");
+
+	    // Crude, only works with single byte ITF8 values
+	    if (m->encoding == E_EXTERNAL ||
+		m->encoding == E_BYTE_ARRAY_STOP ||
+		m->encoding == E_BYTE_ARRAY_LEN) {
+		HashData hd;
+		hd.i = (unsigned char)data[m->offset + m->size-1];
+
+		k = (m->key << 8) | hd.i;
+		HashTableAdd(ds_h, (char *)k, 4, hd, NULL);
+	    }
 	}
     }
 }
@@ -78,15 +90,37 @@ static cram_map *map_find(cram_map **map, unsigned char *key, int id) {
 
 void dump_core_block(cram_block *b, int verbose) {
     int i;
+    int binary = 0, len;
 
-    printf("Data = {");
-    for (i = 0; (verbose || i < 100) && i < b->uncomp_size; i++) {
-	printf(i ? ", %02x" : "%02x", (unsigned char)b->data[i]);
+    len = verbose ? b->uncomp_size : MIN(100, b->uncomp_size);
+    for (i = 0; i < len; i++) {
+	if (isprint(b->data[i]))
+	    continue;
+	binary++;
+    }
+
+    if (binary * 10 > len) {
+	printf("Data = {");
+	for (i = 0; i < len; i++) {
+	    printf(i ? ", %02x" : "%02x", (unsigned char)b->data[i]);
+	}
+
+	if (i < b->uncomp_size)
+	    puts(", ...}");
+	else
+	    puts("}");
+
+    } else {
+	for (i = 0; i < len; i++) {
+	    if (isprint(b->data[i]))
+		putchar(b->data[i]);
+	    else
+		printf("\\%03o", b->data[i]);
+	}
+
+	if (i < b->uncomp_size)
+	    puts("...");
     }
-    if (i < b->uncomp_size)
-	printf(", ...}\n");
-    else
-	printf("}\n");
 }
 
 void dump_seq_block(cram_block *b, int verbose) {
@@ -102,7 +136,10 @@ void dump_seq_block(cram_block *b, int verbose) {
 void dump_quality_block(cram_block *b, int verbose) {
     int i;
     for (i = 0; (verbose || i < 100) && i < b->uncomp_size; i++) {
-	putchar(b->data[i] + '!');
+	if (isprint(b->data[i] + '!'))
+	    putchar(b->data[i] + '!');
+	else
+	    printf("\\%03o", b->data[i] + '!');
     }
     putchar('\n');
 }
@@ -117,10 +154,6 @@ void dump_name_block(cram_block *b, int verbose) {
     }
 }
 
-void dump_mate_info_block(cram_block *b, int verbose) {
-    return dump_core_block(b, verbose);
-}
-
 void dump_tag_block(cram_block *b, int verbose) {
     return dump_core_block(b, verbose);
 }
@@ -130,8 +163,16 @@ int main(int argc, char **argv) {
     cram_container *c;
     off_t pos, pos2, hpos;
     int verbose = 0;
+    HashTable *bsize_h;
+    HashTable *ds_h; // content_id to data-series lookup.
 
-    static int64_t bsize[100], bmax = 0;
+    static bmax = 0;
+    bsize_h = HashTableCreate(128, HASH_DYNAMIC_SIZE|
+			    HASH_NONVOLATILE_KEYS |
+			    HASH_INT_KEYS);
+    ds_h = HashTableCreate(128, HASH_DYNAMIC_SIZE|
+			   HASH_NONVOLATILE_KEYS |
+			   HASH_INT_KEYS);
 
     if (argc >= 2 && strcmp(argv[1], "-v") == 0) {
 	argc--;
@@ -209,11 +250,11 @@ int main(int argc, char **argv) {
 
 	printf("\n      Record encoding map:\n");
 	DumpMap2(c->comp_hdr->rec_encoding_map, stdout, "\t", 
-		 (char *)c->comp_hdr_block->data);
+		 (char *)c->comp_hdr_block->data, ds_h);
 
 	printf("\n      Tag encoding map:\n");
 	DumpMap2(c->comp_hdr->tag_encoding_map, stdout, "\t",
-		 (char *)c->comp_hdr_block->data);
+		 (char *)c->comp_hdr_block->data, ds_h);
 
 	for (j = 0; j < c->num_landmarks; j++) {
 	    cram_slice *s;
@@ -249,10 +290,23 @@ int main(int argc, char **argv) {
 		printf("\tRef base id:     %d\n", s->hdr->ref_base_id);
 	    }
 	
-	    for (id = 0; id < s->hdr->num_blocks; id++)
-		bsize[id] += s->block[id]->comp_size;
-	    if (bmax < s->hdr->num_blocks)
-		bmax = s->hdr->num_blocks;
+	    for (id = 0; id < s->hdr->num_blocks; id++) {
+		HashItem *hi;
+		int k = s->block[id]->content_type == CORE
+		    ? -1 : s->block[id]->content_id;
+		hi = HashTableSearch(bsize_h, (char *)k, 4);
+		if (hi) {
+		    hi->data.i += s->block[id]->comp_size;
+		} else {
+		    HashData hd;
+		    hd.i = s->block[id]->comp_size;
+		    HashTableAdd(bsize_h, (char *)k, 4, hd, NULL);
+		}
+
+		// WARNING: scuppered by having high content_id values.
+		if (bmax < s->block[id]->content_id)
+		    bmax = s->block[id]->content_id;
+	    }
 
 	    for (id = 0; id < s->hdr->num_blocks; id++)
 		cram_uncompress_block(s->block[id]);
@@ -274,39 +328,39 @@ int main(int argc, char **argv) {
 			   (int)b->byte, b->bit);
 
 		    out_sz = 1; /* decode 1 item */
-		    r = c->comp_hdr->BF_codec->decode(s,c->comp_hdr->BF_codec, b, (char *)&bf, &out_sz);
+		    r = c->comp_hdr->codecs[DS_BF]->decode(s,c->comp_hdr->codecs[DS_BF], b, (char *)&bf, &out_sz);
 		    printf("BF = %d => SAM 0x%x (ret %d, out_sz %d)\n", bf, fd->bam_flag_swap[bf], r, out_sz);
 		    bf = fd->bam_flag_swap[bf];
 
 		    if (!IS_CRAM_1_VERS(fd)) {
-			r = c->comp_hdr->CF_codec->decode(s,c->comp_hdr->CF_codec, b, (char *)&i32, &out_sz);
+			r = c->comp_hdr->codecs[DS_CF]->decode(s,c->comp_hdr->codecs[DS_CF], b, (char *)&i32, &out_sz);
 			printf("CF = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
 			cf = i32;
 		    } else {
-			r = c->comp_hdr->CF_codec->decode(s,c->comp_hdr->CF_codec, b, (char *)&cf, &out_sz);
+			r = c->comp_hdr->codecs[DS_CF]->decode(s,c->comp_hdr->codecs[DS_CF], b, (char *)&cf, &out_sz);
 			printf("CF = %d (ret %d, out_sz %d)\n", cf, r, out_sz);
 		    }
 
 		    if (!IS_CRAM_1_VERS(fd) && s->hdr->ref_seq_id == -2) {
 			int32_t ri;
-			r |= c->comp_hdr->RI_codec->decode(s, c->comp_hdr->RI_codec, b, (char *)&ri, &out_sz);
+			r |= c->comp_hdr->codecs[DS_RI]->decode(s, c->comp_hdr->codecs[DS_RI], b, (char *)&ri, &out_sz);
 			printf("RI = %d (ret %d, out_sz %d)\n", ri, r, out_sz);
 		    }
 
-		    r = c->comp_hdr->RL_codec->decode(s,c->comp_hdr->RL_codec, b, (char *)&rl, &out_sz);
+		    r = c->comp_hdr->codecs[DS_RL]->decode(s,c->comp_hdr->codecs[DS_RL], b, (char *)&rl, &out_sz);
 		    printf("RL = %d (ret %d, out_sz %d)\n", rl, r, out_sz);
 
-		    r = c->comp_hdr->AP_codec->decode(s,c->comp_hdr->AP_codec, b, (char *)&i32, &out_sz);
+		    r = c->comp_hdr->codecs[DS_AP]->decode(s,c->comp_hdr->codecs[DS_AP], b, (char *)&i32, &out_sz);
 		    printf("AP = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
 
-		    r = c->comp_hdr->RG_codec->decode(s,c->comp_hdr->RG_codec, b, (char *)&i32, &out_sz);
+		    r = c->comp_hdr->codecs[DS_RG]->decode(s,c->comp_hdr->codecs[DS_RG], b, (char *)&i32, &out_sz);
 		    printf("RG = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
 
 		    if (c->comp_hdr->read_names_included) {
 			int32_t out_sz2 = 1;
 			cram_block *dat = cram_new_block(EXTERNAL, 0);
 
-			r = c->comp_hdr->RN_codec->decode(s,c->comp_hdr->RN_codec, b, (char *)dat, &out_sz2);
+			r = c->comp_hdr->codecs[DS_RN]->decode(s,c->comp_hdr->codecs[DS_RN], b, (char *)dat, &out_sz2);
 			printf("RN = %.*s (ret %d, out_sz %d)\n", out_sz2, BLOCK_DATA(dat), r, out_sz2);
 			cram_free_block(dat);
 		    }
@@ -316,37 +370,37 @@ int main(int argc, char **argv) {
 			puts("Detached");
 			/* MF, RN if !captureReadNames, NS, NP, IS */
 			if (!IS_CRAM_1_VERS(fd)) {
-			    r = c->comp_hdr->MF_codec->decode(s,c->comp_hdr->MF_codec, b, (char *)&i32, &out_sz);
+			    r = c->comp_hdr->codecs[DS_MF]->decode(s,c->comp_hdr->codecs[DS_MF], b, (char *)&i32, &out_sz);
 			    printf("MF = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
 			} else {
-			    r = c->comp_hdr->MF_codec->decode(s,c->comp_hdr->MF_codec, b, &mf, &out_sz);
+			    r = c->comp_hdr->codecs[DS_MF]->decode(s,c->comp_hdr->codecs[DS_MF], b, &mf, &out_sz);
 			    printf("MF = %d (ret %d, out_sz %d)\n", mf, r, out_sz);
 			}
 			if (!c->comp_hdr->read_names_included) {
 			    cram_block *dat = cram_new_block(EXTERNAL, 0);
 			    int32_t out_sz2 = 1;
 
-			    r = c->comp_hdr->RN_codec->decode(s,c->comp_hdr->RN_codec, b, (char *)dat, &out_sz2);
+			    r = c->comp_hdr->codecs[DS_RN]->decode(s,c->comp_hdr->codecs[DS_RN], b, (char *)dat, &out_sz2);
 			    printf("RN = %.*s (ret %d, out_sz %d)\n", out_sz2, BLOCK_DATA(dat), r, out_sz2);
 			    cram_free_block(dat);
 			}
 		    
-			r = c->comp_hdr->NS_codec->decode(s,c->comp_hdr->NS_codec, b, (char *)&i32, &out_sz);
+			r = c->comp_hdr->codecs[DS_NS]->decode(s,c->comp_hdr->codecs[DS_NS], b, (char *)&i32, &out_sz);
 			printf("NS = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
 
-			r = c->comp_hdr->NP_codec->decode(s,c->comp_hdr->NP_codec, b, (char *)&i32, &out_sz);
+			r = c->comp_hdr->codecs[DS_NP]->decode(s,c->comp_hdr->codecs[DS_NP], b, (char *)&i32, &out_sz);
 			printf("NP = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
 
-			r = c->comp_hdr->TS_codec->decode(s,c->comp_hdr->TS_codec, b, (char *)&i32, &out_sz);
+			r = c->comp_hdr->codecs[DS_TS]->decode(s,c->comp_hdr->codecs[DS_TS], b, (char *)&i32, &out_sz);
 			printf("TS = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
 		    } else if (cf & CRAM_FLAG_MATE_DOWNSTREAM) {
 			puts("Not detached, and mate is downstream");
-			r = c->comp_hdr->NF_codec->decode(s,c->comp_hdr->NF_codec, b, (char *)&i32, &out_sz);
+			r = c->comp_hdr->codecs[DS_NF]->decode(s,c->comp_hdr->codecs[DS_NF], b, (char *)&i32, &out_sz);
 			printf("NF = %d+%d = %d (ret %d, out_sz %d)\n", i32, rec+1, i32+rec+1, r, out_sz);
 		    }
 
 		    if (IS_CRAM_1_VERS(fd)) {
-			r = c->comp_hdr->TC_codec->decode(s,c->comp_hdr->TC_codec, b, (char *)&ntags, &out_sz);
+			r = c->comp_hdr->codecs[DS_TC]->decode(s,c->comp_hdr->codecs[DS_TC], b, (char *)&ntags, &out_sz);
 			printf("TC = %d (ret %d, out_sz %d)\n", ntags, r, out_sz);
 
 			for (f = 0; f < ntags; f++) {
@@ -355,7 +409,7 @@ int main(int argc, char **argv) {
 			    cram_map *m;
 			    cram_block *tag = cram_new_block(EXTERNAL, 0);
 
-			    r = c->comp_hdr->TN_codec->decode(s, c->comp_hdr->TN_codec,
+			    r = c->comp_hdr->codecs[DS_TN]->decode(s, c->comp_hdr->codecs[DS_TN],
 							      b, (char *)&id, &out_sz);
 			    key[0] = (id>>16)&0xff;
 			    key[1] = (id>>8)&0xff;
@@ -385,7 +439,7 @@ int main(int argc, char **argv) {
 		    } else {
 			int32_t tl, ntags;
 			char *tn;
-			r = c->comp_hdr->TL_codec->decode(s,c->comp_hdr->TL_codec, b, (char *)&tl, &out_sz);
+			r = c->comp_hdr->codecs[DS_TL]->decode(s,c->comp_hdr->codecs[DS_TL], b, (char *)&tl, &out_sz);
 			printf("TL = %d (ret %d, out_sz %d)\n", tl, r, out_sz);
 
 			tn = (char *)c->comp_hdr->TL[tl];
@@ -426,7 +480,7 @@ int main(int argc, char **argv) {
 		    }
 
 		    if (!(bf & BAM_FUNMAP)) {
-			r = c->comp_hdr->FN_codec->decode(s,c->comp_hdr->FN_codec, b, (char *)&fn, &out_sz);
+			r = c->comp_hdr->codecs[DS_FN]->decode(s,c->comp_hdr->codecs[DS_FN], b, (char *)&fn, &out_sz);
 			printf("FN = %d (ret %d, out_sz %d)\n", fn, r, out_sz);
 
 			prev_pos = 0;
@@ -434,10 +488,10 @@ int main(int argc, char **argv) {
 			    char op;
 			    int32_t pos;
 
-			    r = c->comp_hdr->FC_codec->decode(s,c->comp_hdr->FC_codec, b, &op, &out_sz);
+			    r = c->comp_hdr->codecs[DS_FC]->decode(s,c->comp_hdr->codecs[DS_FC], b, &op, &out_sz);
 			    printf("  %d: FC = %c (ret %d, out_sz %d)\n", f, op, r, out_sz);
 
-			    r = c->comp_hdr->FP_codec->decode(s,c->comp_hdr->FP_codec, b, (char *)&pos, &out_sz);
+			    r = c->comp_hdr->codecs[DS_FP]->decode(s,c->comp_hdr->codecs[DS_FP], b, (char *)&pos, &out_sz);
 			    printf("  %d: FP = %d+%d = %d (ret %d, out_sz %d)\n", f, pos, prev_pos, pos + prev_pos, r, out_sz);
 
 			    pos += prev_pos;
@@ -450,13 +504,13 @@ int main(int argc, char **argv) {
 
 				dat[0]='?';dat[1]=0;
 				if (IS_CRAM_1_VERS(fd)) {
-				    if (c->comp_hdr->IN_codec) {
-					r = c->comp_hdr->IN_codec->decode(s,c->comp_hdr->IN_codec, b, dat, &out_sz2);
+				    if (c->comp_hdr->codecs[DS_IN]) {
+					r = c->comp_hdr->codecs[DS_IN]->decode(s,c->comp_hdr->codecs[DS_IN], b, dat, &out_sz2);
 					printf("  %d: IN(S) = %.*s (ret %d, out_sz %d)\n", f, out_sz2, dat, r, out_sz2);
 				    }
 				} else {
-				    if (c->comp_hdr->SC_codec) {
-					r = c->comp_hdr->SC_codec->decode(s,c->comp_hdr->SC_codec, b, dat, &out_sz2);
+				    if (c->comp_hdr->codecs[DS_SC]) {
+					r = c->comp_hdr->codecs[DS_SC]->decode(s,c->comp_hdr->codecs[DS_SC], b, dat, &out_sz2);
 					printf("  %d: SC(S) = %.*s (ret %d, out_sz %d)\n", f, out_sz2, dat, r, out_sz2);
 				    }
 				}
@@ -465,13 +519,13 @@ int main(int argc, char **argv) {
 
 			    case 'X': { // Substitution; BS
 				char bs;
-				r = c->comp_hdr->BS_codec->decode(s,c->comp_hdr->BS_codec, b, &bs, &out_sz);
+				r = c->comp_hdr->codecs[DS_BS]->decode(s,c->comp_hdr->codecs[DS_BS], b, &bs, &out_sz);
 				printf("  %d: BS = %d (ret %d)\n", f, bs, r);
 				break;
 			    }
 
 			    case 'D': { // Deletion; DL
-				r = c->comp_hdr->DL_codec->decode(s,c->comp_hdr->DL_codec, b, (char *)&i32, &out_sz);
+				r = c->comp_hdr->codecs[DS_DL]->decode(s,c->comp_hdr->codecs[DS_DL], b, (char *)&i32, &out_sz);
 				printf("  %d: DL = %d (ret %d)\n", f, i32, r);
 				break;
 			    }
@@ -481,45 +535,45 @@ int main(int argc, char **argv) {
 				int32_t out_sz2 = 1;
 
 				dat[0]='?';dat[1]=0;
-				r = c->comp_hdr->IN_codec->decode(s,c->comp_hdr->IN_codec, b, dat, &out_sz2);
+				r = c->comp_hdr->codecs[DS_IN]->decode(s,c->comp_hdr->codecs[DS_IN], b, dat, &out_sz2);
 				printf("  %d: IN(I) = %.*s (ret %d, out_sz %d)\n", f, out_sz2, dat, r, out_sz2);
 				break;
 			    }
 
 			    case 'i': { // Insertion (single base); BA
 				char cc;
-				r = c->comp_hdr->BA_codec->decode(s,c->comp_hdr->BA_codec, b, &cc, &out_sz);
+				r = c->comp_hdr->codecs[DS_BA]->decode(s,c->comp_hdr->codecs[DS_BA], b, &cc, &out_sz);
 				printf("  %d: BA = %c (ret %d)\n", f, cc, r);
 				break;
 			    }
 
 			    case 'B': { // Read base; BA, QS
 				char cc, qc;
-				r  = c->comp_hdr->BA_codec->decode(s,c->comp_hdr->BA_codec, b, &cc, &out_sz);
-				r |= c->comp_hdr->QS_codec->decode(s,c->comp_hdr->QS_codec, b, &qc, &out_sz);
+				r  = c->comp_hdr->codecs[DS_BA]->decode(s,c->comp_hdr->codecs[DS_BA], b, &cc, &out_sz);
+				r |= c->comp_hdr->codecs[DS_QS]->decode(s,c->comp_hdr->codecs[DS_QS], b, &qc, &out_sz);
 				printf("  %d: BA/QS(B) = %c/%d (ret %d)\n", f, cc, qc, r);
 				break;
 			    }
 
 			    case 'Q': { // Quality score; QS
 				char qc;
-				r = c->comp_hdr->QS_codec->decode(s,c->comp_hdr->QS_codec, b, &qc, &out_sz);
+				r = c->comp_hdr->codecs[DS_QS]->decode(s,c->comp_hdr->codecs[DS_QS], b, &qc, &out_sz);
 				printf("  %d: QS = %d (ret %d)\n", f, qc, r);
 				break;
 			    }
 
 			    case 'N':
-				r = c->comp_hdr->RS_codec->decode(s,c->comp_hdr->RS_codec, b, (char *)&i32, &out_sz);
+				r = c->comp_hdr->codecs[DS_RS]->decode(s,c->comp_hdr->codecs[DS_RS], b, (char *)&i32, &out_sz);
 				printf("  %d: RS = %d (ret %d)\n", f, i32, r);
 				break;
 
 			    case 'P':
-				r = c->comp_hdr->PD_codec->decode(s,c->comp_hdr->PD_codec, b, (char *)&i32, &out_sz);
+				r = c->comp_hdr->codecs[DS_PD]->decode(s,c->comp_hdr->codecs[DS_PD], b, (char *)&i32, &out_sz);
 				printf("  %d: PD = %d (ret %d)\n", f, i32, r);
 				break;
 
 			    case 'H':
-				r = c->comp_hdr->HC_codec->decode(s,c->comp_hdr->HC_codec, b, (char *)&i32, &out_sz);
+				r = c->comp_hdr->codecs[DS_HC]->decode(s,c->comp_hdr->codecs[DS_HC], b, (char *)&i32, &out_sz);
 				printf("  %d: HC = %d (ret %d)\n", f, i32, r);
 				break;
 
@@ -528,7 +582,7 @@ int main(int argc, char **argv) {
 			    }
 			}
 
-			r = c->comp_hdr->MQ_codec->decode(s,c->comp_hdr->MQ_codec, b, (char *)&i32, &out_sz);
+			r = c->comp_hdr->codecs[DS_MQ]->decode(s,c->comp_hdr->codecs[DS_MQ], b, (char *)&i32, &out_sz);
 			printf("MQ = %d (ret %d, out_sz %d)\n", i32, r, out_sz);
 
 			if (cf & CRAM_FLAG_PRESERVE_QUAL_SCORES) {
@@ -536,7 +590,7 @@ int main(int argc, char **argv) {
 			    int32_t out_sz2 = rl, i;
 
 			    dat[0]='?';dat[1]=0;
-			    r = c->comp_hdr->Qs_codec->decode(s,c->comp_hdr->Qs_codec, b, dat, &out_sz2);
+			    r = c->comp_hdr->codecs[DS_Qs]->decode(s,c->comp_hdr->codecs[DS_Qs], b, dat, &out_sz2);
 			    for (i = 0; i < rl; i++)
 				dat[i] += '!';
 			    printf("Qs = %.*s (ret %d, out_sz %d)\n", out_sz2, dat, r, out_sz2);
@@ -548,7 +602,7 @@ int main(int argc, char **argv) {
 
 			do {
 			    int32_t out_sz2 = len > 1024 ? 1024 : len;
-			    r = c->comp_hdr->BA_codec->decode(s, c->comp_hdr->BA_codec, b, dat, &out_sz2);
+			    r = c->comp_hdr->codecs[DS_BA]->decode(s, c->comp_hdr->codecs[DS_BA], b, dat, &out_sz2);
 			    printf("SQ = %.*s (out_sz %d)\n", out_sz2, dat, out_sz2);
 			    len -= 1024;
 			} while (len > 0);
@@ -558,7 +612,7 @@ int main(int argc, char **argv) {
 
 			    do {
 				int32_t out_sz2 = len > 1024 ? 1024 : len;
-				r = c->comp_hdr->Qs_codec->decode(s, c->comp_hdr->Qs_codec, b, dat, &out_sz2);
+				r = c->comp_hdr->codecs[DS_Qs]->decode(s, c->comp_hdr->codecs[DS_Qs], b, dat, &out_sz2);
 				for (i = 0; i < len; i++)
 				    dat[i] += '!';
 				printf("Qs = %.*s (out_sz %d)\n", out_sz2, dat, out_sz2);
@@ -586,31 +640,52 @@ int main(int argc, char **argv) {
 		if (b->content_type == CORE) {
 		    dump_core_block(b, verbose);
 		} else {
-		    switch (b->content_id) {
-		    case 0:
-			dump_seq_block(b, verbose);
-			break;
+		    int t, m;
+		    enum cram_fields cf = 0;
+		    char fields[1024], *fp = fields;
+		    for (m = 0; m < 2; m++) {
+			cram_map **ma = m
+			    ? c->comp_hdr->tag_encoding_map
+			    : c->comp_hdr->rec_encoding_map;
 			
-		    case 1:
-			dump_quality_block(b, verbose);
-			break;
-			
-		    case 2:
-			dump_name_block(b, verbose);
-			break;
-			
-		    case 3:
-			dump_mate_info_block(b, verbose);
-			break;
-			
-		    case 4:
-			dump_tag_block(b, verbose);
-			break;
+			for (t = 0; t < CRAM_MAP_HASH; t++) {
+			    cram_map *m;
+			    unsigned char *data = c->comp_hdr_block->data;
+			    for (m = ma[t]; m; m = m->next) {
+				if (m->encoding != E_EXTERNAL &&
+				    m->encoding != E_BYTE_ARRAY_STOP &&
+				    m->encoding != E_BYTE_ARRAY_LEN)
+				    continue;
+				if (data[m->offset + m->size-1] !=
+				    b->content_id)
+				    continue;
+				if ((m->key>>16)&0xff)
+				    *fp++ = (m->key>>16)&0xff;
+				*fp++ = (m->key>> 8)&0xff;
+				*fp++ = (m->key>> 0)&0xff;
+				*fp++ = ' ';
 
-		    default:
-			dump_core_block(b, verbose); // hex
-			break;
+				if (m->key>>16)             cf |= CRAM_aux;
+				if (m->key == ('Q'<<8)+'S') cf |= CRAM_QS;
+				if (m->key == ('R'<<8)+'N') cf |= CRAM_RN;
+				if (m->key == ('S'<<8)+'C') cf |= CRAM_SC;
+				if (m->key == ('I'<<8)+'N') cf |= CRAM_IN;
+			    }
+			}
 		    }
+		    *fp++ = 0;
+		    printf("\t    Keys:         %s\n", fields);
+
+		    if (cf == CRAM_aux)
+			dump_tag_block(b, verbose);
+		    else if (cf == CRAM_QS)
+			dump_quality_block(b, verbose);
+		    else if (cf == CRAM_SC)
+			dump_seq_block(b, verbose);
+		    else if (cf == CRAM_RN)
+			dump_name_block(b, verbose);
+		    else
+			dump_core_block(b, verbose);
 		}
 	    }
 
@@ -627,10 +702,50 @@ int main(int argc, char **argv) {
 
     {
 	int id;
+
 	puts("");
-	for (id = 0; id < bmax; id++)
-	    printf("Block %d, total size %"PRId64"\n", id, bsize[id]);
+	for (id = -1; id <= bmax; id++) {
+	    int k;
+	    HashItem *hi;
+	    HashIter *iter;
+
+	    if (!(hi = HashTableSearch(bsize_h, (char *)id, 4)))
+		continue;
+
+	    k = (int)hi->key;
+	    if (k == -1) {
+		printf("Block CORE          , total size %10ld\n", hi->data.i);
+		continue;
+	    }
+
+	    printf("Block content_id %3d, total size %10ld ", k, hi->data.i);
+
+	    iter = HashTableIterCreate();
+	    while ((hi = HashTableIterNext(ds_h, iter))) {
+		int c;
+		char buf[5];
+		int x = 4;
+
+		if (hi->data.i != k)
+		    continue;
+		
+		c = ((int)hi->key)>>8;
+		
+		buf[x--] = 0;
+		while(c & 0xff) {
+		    buf[x--] = c;
+		    c >>= 8;
+		}
+		printf(" %s", &buf[x+1]);
+	    }
+	    putchar('\n');
+
+	    HashTableIterDestroy(iter);
+	}
     }
 
+    HashTableDestroy(bsize_h, 0);
+    HashTableDestroy(ds_h, 0);
+
     return 0;
 }
